"""Module containing various plotting constants shared between modules."""

NUMERICAL_CMAPS = ('viridis', 'magma', 'plasma', 'inferno')

CATEGORICAL_COLORS = ('#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
                      '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf')
